# Copyright(c) 2001-2003 Association of Universities for Research in Astronomy, Inc.
procedure wmef(input,output)

# WMEF creates a MEF file from input FITS or PLIO (.pl) images.
# Input can also be extensions in already existing MEF files.
# There is no attempt to handle other image types.
#
# Version: Aug 17, 2001 MT,BM,IJ v1.2 release
#          Jan 30, 2002 IJ       removed gsetsec, calling task needs to take care of this
#          Feb 28, 2002 IJ       v1.3 release
#          Sept 20,2002 IJ       v1.4 release
#          Jan 10, 2003 IJ       fixed ad+ -> add+ bug
#          Aug 15, 2003 KL       Port to IRAF2.12.  hedit: addonly

char    input    {prompt="Input images in the order they should appear"}
char    output   {prompt="Output multi-extension fits file"}
char    extnames {"SCI,VAR,DQ",prompt="Extension names (optional)"}
char    phu      {"dummy",prompt="Source of PHU info"}
bool    verbose  {no,prompt="Verbose"}
int     status   {0,prompt="Exit status (0=good)"}

begin

char l_input, l_output, l_extnames, l_phu
struct datestamp
int    ii, jj, n_images, n_match, n_fits, n_extnames, n_inext
bool  l_verbose

char  l_listim, l_delim, l_im
char  tmpin, tmpfiles, tmpext, tmpfits

tmpin = mktemp("tmpin")
tmpfiles = mktemp("tmpfil")
tmpext = mktemp("tmpext")
tmpfits = mktemp("tmpfit")
  
status=0
# Query parameters
l_input = input ; l_output = output ; l_extnames = extnames
l_phu = phu ; l_verbose=verbose

cache("imgets","tinfo")

# check access to phu if not "dummy" or "none"
if(l_phu!="dummy" && l_phu!="none" && l_phu!="" && l_phu!=" ") {
  if(substr(l_phu,strlen(l_phu)-7,strlen(l_phu))!=".fits[0]") {
    if(substr(l_phu,strlen(l_phu)-4,strlen(l_phu))==".fits") {
      l_phu=l_phu//"[0]"
    } else if (substr(l_phu,strlen(l_phu)-2,strlen(l_phu))=="[0]")
      l_phu=substr(l_phu,1,strlen(l_phu)-3)//".fits[0]"
    else
      l_phu=l_phu//".fits[0]"
  }

  if(!imaccess(l_phu)) {
    print("ERROR - WMEF: PHU source "//l_phu//" does not exist")
    status=1
    bye
  }
}

# check input 
if(l_input=="" || l_input==" ") {
  print("ERROR - WMEF: Input images are specified as an empty string")
  status=1
  bye
}

# check access to output
if(l_output=="" || l_output==" ") {
  print("ERROR - WMEF: Output MEF file is specified as an empty string")
  status=1
  bye
}
if(imaccess(l_output)) {
  print("ERROR - WMEF: Output MEF file "//l_output//" exists")
  status=1
  bye
}
if(stridx("[",l_output)!=0) {
  print("ERROR - WMEF: Output MEF file "//l_output//" should be defined without extension number")
  status=1
  bye
}
if(substr(l_output,(strlen(l_output)-4),strlen(l_output))!=".fits") 
  l_output=l_output//".fits"

# Cheap and not quite reliable check for image sections
print(l_input) | match(":","STDIN",stop-,print-,meta+) | count("STDIN") | scan(n_match)
if(n_match>0) {
  print("ERROR - WMEF: Input files "//l_input//" contain image sections (colons)")
  status=1
  bye
}

# Make lists of input and extensions - sed cannot read reliably from
# a pipe in cl, but using sed is more efficient than stepping through
# this list looking for commas
print(l_input, > tmpin)
sed("-e",'s/\,/\ /g',tmpin) | words("STDIN", > tmpfiles)
delete(tmpin,ver-, >>& "dev$null")
n_extnames=0
if(l_extnames!="" && l_extnames!=" ") {
  print(l_extnames, > tmpin)
  sed("-e",'s/\,/\ /g',tmpin) | words("STDIN", > tmpext)
  delete(tmpin,ver-, >>& "dev$null")
  count(tmpext) | scan(n_extnames)
}
count(tmpfiles) | scan(n_images)
if(n_images!=n_extnames && n_extnames!=0) {
  n_extnames=0
  delete(tmpext,ver-, >>& "dev$null")
  print("WARNING - WMEF: Number of input images and extensions names differ.")
  print("                Extensions are not named.")
}

# Check access and format for input images. 
for(ii=1;ii<=n_images;ii+=1) {
  fields(tmpfiles,"1",lines=str(ii),quit_if_miss-,print_file_n-) | scan(l_im)
  if(!imaccess(l_im)) {
    print("ERROR - WMEF: Input image "//l_im//" not found.")
    status=1
    delete(tmpfiles//","//tmpext,ver-, >>& "dev$null")
    bye
  }

# make sure that input is not a full MEF file with no extension
# defined
if(stridx("[",l_im)==0 && access(l_im//".fits") || substr(l_im,strlen(l_im)-4,strlen(l_im))==".fits") {
  imgets(l_im//"[0]","EXTEND", >>& "dev$null")
    if(imgets.value=="T" && imaccess(l_im//"[1]")) {
      print("ERROR - WMEF: The image "//l_im//" is a MEF file without specified extension.")
      print("              wmef needs the extension to be defined.")
      status=1
      delete(tmpfiles//","//tmpext,ver-, >>& "dev$null")
      bye
    } else {
	l_im=l_im//"[0]"
    }
}

# Make sure that input is either FITS or .pl
# It is not a requirement that there is a pixel file, since we want
# to be able to attach bintables with this task at some later point.

# FITS bintables also has i_pixfile defined ...
imhead(l_im,long+,userf+) | match("Pixel file") | match(".fits") | \
  count("STDIN") | scan(n_fits)
if(n_fits!=1) {
  if(!access(l_im//".pl") && substr(l_im,strlen(l_im)-2,strlen(l_im))!=".pl") {

      print("ERROR - WMEF: Image "//l_im//" is neither FITS nor a PLIO (.pl) file."
      print("              wmef requires FITS or PLIO files.")
      status=1
      delete(tmpfiles//","//tmpext,ver-, >>& "dev$null")
      bye
  }
}
    
} # end of checking the files

# Create .fits from .pl if needed.
# Build the string of fits files to use.
l_listim="" ; l_delim=""
for(ii=1;ii<=n_images;ii+=1) {
  fields(tmpfiles,"1",lines=str(ii),quit_if_miss-,print_file_n-) | scan(l_im)
  if(access(l_im//".pl") || substr(l_im,strlen(l_im)-2,strlen(l_im))==".pl") {
     imarith(l_im,"+","0",tmpfits//str(ii)//".fits",title="",divzero=0.,
     hparams="",pixtype="short",calctype="",verbose-,noact-)
     if(ii==1) {
       l_listim = tmpfits//str(ii)//".fits"
       l_delim=tmpfits//str(ii)//".fits"
     } else {
       l_listim = l_listim//","//tmpfits//str(ii)//".fits"
       l_delim=l_delim//","//tmpfits//str(ii)//".fits"
     }
  } else {

     # make sure .fits is included in the image name
     if(stridx("[",l_im)!=0) {
        if(substr(l_im,(stridx("[",l_im)-5),(stridx("[",l_im)-1) )!=".fits")

          l_im=substr(l_im,1,(stridx("[",l_im)-1))//".fits"//substr(l_im,stridx("[",l_im),strlen(l_im))
     }
     else {
        if(substr(l_im,strlen(l_im)-4,strlen(l_im))!=".fits")
          l_im=l_im//".fits"
     }

     # Check for a table and ensure bintables are accessed ok
     tinfo(l_im,ttout-, >>& "dev$null")
     if (tinfo.tbltype=="fits") { 
	 l_im=l_im//"[1]"

# If extension is attached, make sure reference is by number not name
# since there is a bug in fxinsert
# We know that the extension is last and if it is by name that name
# is in the header as EXTNAME
     } else if(stridx("[",l_im)!=0) {
       imgets(l_im,"EXTNAME")
       if(imgets.value == substr(l_im,stridx("[",l_im)+1,strlen(l_im)-1) && imgets.value!="0" ) {
          # Let's then attempt to find the correct extension number
#          print("By name")
          imgets(substr(l_im,1,stridx("[",l_im)-1)//"[0]","NEXTEND")
          n_inext=int(imgets.value)
          for(jj=1;jj<=n_inext;jj+=1) {
            imgets(substr(l_im,1,stridx("[",l_im)-1)//"["//str(jj)//"]","EXTNAME")
            if(imgets.value==substr(l_im,stridx("[",l_im)+1,strlen(l_im)-1) )
                l_im=substr(l_im,1,stridx("[",l_im)-1)//"["//str(jj)//"]"

          }

       }
     }

     if(ii==1)
       l_listim = l_im
     else
       l_listim = l_listim//","//l_im
  }

} # end of building the strings

if(l_verbose) {
  print("Input files: "//l_listim)
#  print(l_delim)
}

# get or make the PHU

if(l_phu=="dummy" || l_phu=="none" || l_phu=="" || l_phu==" ") 
   fxdummyh(l_output,hdr_file="")
else
   fxcopy(l_phu,l_output,groups="",new_file+,verbose=l_verbose)

# write the whole thing
if(l_verbose)
fxinsert(l_listim,l_output//"[0]",group="",verbose=l_verbose)
else
fxinsert(l_listim,l_output//"[0]",group="",verbose=l_verbose, >>& "dev$null")

# put number of extensions in the header - not sure why that does
# not get done by fxinsert
hedit(l_output//"[0]","NEXTEND",n_images,add+,addonly-,del-,ver-,show-,up+)

# name the extensions 
if(n_extnames!=0) {
  for(ii=1;ii<=n_extnames;ii+=1) {
    fields(tmpext,"1",lines=str(ii),quit_if_miss-,print_file_n-) | scan(l_im)
    tinfo(l_output//"["//str(ii)//"]",ttout-, >>& "dev$null")
    if (tinfo.tbltype=="fits") {
     parkey(l_im,l_output//"["//str(ii)//"]","EXTNAME",add+)
    } else {
     hedit(l_output//"["//str(ii)//"]","extname",l_im,add+,addonly-,del-,
       ver-,show-,up+)
    }
  }
}

# Gemini default header keywords in PHU
date | scan(datestamp)
hedit(l_output//"[0]","WMEF",datestamp,add+,addonly-,del-,ver-,show-,up+)
hedit(l_output//"[0]","GEM-TLM",datestamp,add+,addonly-,del-,ver-,show-,up+)

# Update DATASEC
#gsetsec(l_output,key_datsec = "DATASEC")

# Clean up images and files
imdelete(l_delim,ver-, >>& "dev$null")
delete(tmpfiles//","//tmpext,ver-,  >>& "dev$null")

# set exit status flag
status=0

end
